/*
 * Stochastic Steady State Ecvolutionary algorithm
 * require an extern evaluation function and the gsl libraries included in utilities.h for random number reneraioon
 */

#include <stdio.h>
#include <stdlib.h>

#include <vector>
#include <algorithm>
#include <string.h>
#include <time.h>
#include <typeinfo>
#include "evosss.h"
#include "utilities.h"

int trials=1;
int ttrials=1;

double *bestggenotype = NULL;		// the best generalizing genotype

int popsize = 20;					// popsize (parents)
long int maxevaluations = 100;		// max number of evaluations
int seed = 1;						// random seed
int nreplications = 1;				// number of replications
double prange = 5.0;				// parameters range
double mutrate = 0.03;				// mutation rate
double stochasticity = 0.1;			// stochasticity
int steptimelength = 0;				// time length of a step during graphic visualization
int envchangeevery = 20;			// environment changed every n generations

RandomGenerator* RNG;               // randomgenerator class

char statDir[512];

/*
 * display indfit matrix (in, fitness, noisefitness)
 */
void displayindfit(double **indfit, int popsize)
{
    double **gpi;
    double *in;
    int i,j;
    
    for (i=0, gpi=indfit; i < popsize*2; i++, gpi++)
    {
      for(j=0, in = *gpi; j < 3; j++, in++)
        printf("%.4f ", *in);
      printf("\n");
    }
}

/*
 * sort indfit matrix on the basis of noisefitness in descending order
 * offspring are preferred to parents having the same fitness
 */
void sortindfit(double **indfit, int popsize, double *maxnfitness, double *maxfitness, double *avefitness)
{
    double **gpi;
    double **gpj;
    double *in;
    double *ij;
    int i,j;
    double cmax;
    double fmax;
    double fnmax;
    int    imax;
    
    *avefitness = 0.0;
    for (i=0; i < popsize*2; i++)
       {
        cmax = 0;
        fmax = -9999.0;
        fnmax = -9999.0;
        for (j=0, gpi=indfit; j < popsize*2; j++, gpi++)
          {
            in = *gpi;
            if (*(in + 2) >= fnmax)
              {
                  cmax = *in;
                  fmax = *(in + 1);
                  fnmax = *(in + 2);
                  imax = j;
              }
            if (i == 0)
              *avefitness += *(in + 2);
          }
        if (i == 0)
          {
           *maxfitness = fmax;
           *maxnfitness = fnmax;
          }
       // printf("sort %.4f %.4f %.4f -> %d \n", cmax, fmax, fnmax, popsize*2+i);
        gpi=(indfit + (popsize*2+i));
        in = *gpi;
        *in = cmax; in++;
        *in = fmax; in++;
        *in = fnmax;
        gpi=(indfit + imax);
        in=*gpi;
        *(in + 2) = -9999000.0;
        }
    
     for (i=0, gpi=indfit, gpj = (indfit + (popsize *2)); i < popsize*2; i++, gpi++, gpj++)
        for(j=0, in = *gpi, ij = *gpj; j < 3; j++, in++, ij++)
           *in = *ij;
    
     *avefitness = *avefitness / (double) popsize*2;
}

/*
 * copy and mutate
 */
void copyandmutate(double **genome, int ngenes, double mutr, double prange, int parent, int offspring)
{
    double **gpi;
    double **gpi2;
    double *gi;
    double *gi2;
    int g;
    
    gpi=(genome + parent);
    gpi2=(genome + offspring);
    
    for (g=0,gi = *gpi, gi2 = *gpi2; g < ngenes; g++, gi++, gi2++)
        {
            if ((double) (rand()/(double)RAND_MAX) < mutr)
			 {
			   //printf("%d ", g);
               if (RNG->getDouble(0.0, 1.0) < 0.5)
			     {
                  //*gi2 = (double) rand()/(double) (RAND_MAX * (prange * 2.0) - prange);
                  *gi2 = RNG->getDouble(-prange, prange);
			     }
			     else
			     {
                  //*gi2 += gaussrand();
                  *gi2 += RNG->getGaussian(1.0, 0.0);
				  if (*gi2 > prange)
				     *gi2 = prange;
				  if (*gi2 < -prange)
				     *gi2 = -prange;
				 }
			 }
             else
			 {
               *gi2 = *gi;
			 }
            
        }
//printf("\n");
    //printf("copy and mutate %d %d\n", parent, offspring);
}


/*
 * stochastic steady state
 */
void evolve(int ngenes)
{

    double **genome;
    double **indfit;
    double *bestggeno;
    double **gpi;
    double **ipi;
    double **ipi2;
    double *gi;
	double *gi2;
    double *in;
    double *in2;
    int ceval;
    int cgen;
    double maxnf;
    double maxf;
    double avef;
	double maxgf;
    double f;
    int i;
    int j;
    int ci;
    int s, ss;
    char filen[256];
    FILE *fp;
    double *statistics;                     // performance across generations
    double *stat;                           // performance across generations, local pointer
    long int statn;                         // number of stat data
	double mutr;							// mutation rate (reduced to 0.01 during last phase)
	double stoch;							// stochasticity (reduced to 0.0 during last phase)
	char filename[512];

    printf("SSS: seed %d MaxEval %ld popsize %d prange %.1f mutrate %.1f stochasticity %.1lf\n", seed, maxevaluations, popsize, prange, mutrate, stochasticity);

	// initialize local mutrate and stochasticity
	mutr = mutrate;
	stoch = stochasticity;
	
    // set the seed and generate the initial states
    RNG->setSeed(seed);
    createInitialStates(0);

    // allocate memory for statistics
    statn = (maxevaluations / (trials * popsize) * 7) + 100;
    statistics = (double *) malloc(statn * sizeof(double));
    stat = statistics;
    statn = 0;
	// initialize fitness statistics
	maxgf = 0.0;
    
    // allocate memory for the best genotype
    bestggeno = (double *)malloc(ngenes * sizeof(double));
	
    // allocate and initialize genome (parent, offspring, and best generalizing ind to date)
    genome = (double **) malloc(popsize*2 * sizeof(double *));
    for (i=0, gpi=genome; i < popsize*2; i++,gpi++)
    {
        *gpi = (double *)malloc(ngenes * sizeof(double));
        for (j=0,gi = *gpi; j < ngenes; j++,gi++)
        {
            //*gi = (double) rand()/(double)RAND_MAX * (prange * 2.0) - prange;
            *gi = RNG->getDouble(-prange, prange);
        }
    }
	

    // allocate and initialize indfit matrix with id, fitness, and noisefitness
    indfit = (double **) malloc(popsize*4 * sizeof(double *));
    for (i=0, gpi=indfit; i < popsize*4; i++, gpi++)
    {
        *gpi = (double *) malloc(3 * sizeof(double));
        gi = *gpi;
        *gi = (double) i; gi++;
        *gi = -9999; gi++;
        *gi = -9999;
    }

    printf("Seed %d\n", seed);

    //evaluate parents
    ceval = 0;
    cgen = 0;
    for(i=0, gpi=genome, ipi=indfit; i < popsize; i++, gpi++, ipi++)
      {
          f = evaluate(*gpi, ngenes, 0);
          in = *ipi; in++;
          *in = f; in++;
          *in = f + (double) rand()/(double)RAND_MAX * (stoch * f * 2.0) - (stoch * f);
          ceval += trials;
      }
	
	
    sortindfit(indfit, popsize, &maxnf, &maxf, &avef);
    //displayindfit(indfit, popsize);
    printf("Gen %d Maxn %.4f Max %.4f\n", cgen, maxnf, maxf);
    while (ceval <= maxevaluations)
      {
	  
		 cgen += 1;
	     // during last phase reduce mutation rate and seto stochasticity to null
		if (ceval > (maxevaluations / 20 * 19))
		  {
			stoch = 0.0;
			mutr = 0.01;
		  }
		 // change environment every n generations and re-evaluate parents
         if (cgen > 0 && (cgen % envchangeevery) == 0)
		   {
		    createInitialStates(1);
			printf("Environment varied\n");
		    for(i=0, gpi=genome, ipi=indfit; i < popsize; i++, gpi++, ipi++)
             {
               f = evaluate(*gpi, ngenes, 0);
               in = *ipi; in++;
               *in = f; in++;
               *in = f + (double) rand()/(double)RAND_MAX * (stoch * f * 2.0) - (stoch * f);
               ceval += trials;
             }
		   }
		 
          // regenerate noisefitness
         for(i=0, ipi=indfit; i < popsize; i++, ipi++)
          {
              in = *ipi;
              in++;
              f = *in;
              in++;
              *in = f + (double) rand()/(double)RAND_MAX * (stoch * f * 2.0) - (stoch * f);
          }
         // evaluate offspring
         for(i=0, ipi=(indfit + popsize); i < popsize; i++, ipi++)
          {
              in = *ipi;
              ci = (int) *in;
              gpi=(genome + ci);
              in++;
              *in = f = evaluate(*gpi, ngenes, 0);
              in++;
              *in = f + (double) rand()/(double)RAND_MAX * (stoch * f * 2.0) - (stoch * f);
              ceval += trials;
          }
         // sort indexes by fitness
         sortindfit(indfit, popsize, &maxnf, &maxf, &avef);
		 // test generalization ability of the best individual and store the best found to far
		 ipi=indfit;
		 in = *ipi;
		 ci = (int) *in;
		 gpi=(genome + ci);
		 f = evaluate(*gpi, ngenes, 1);
		 ceval += ttrials;
		 if (f > maxgf)
		   {
		     maxgf = f;
			 for (j=0,gi = *gpi, gi2 = bestggeno; j < ngenes; j++,gi++, gi2++)
			     *gi2 = *gi;
		   }
		 // display statistics
		 printf("Gen %d Maxn %.2f Max %.2f g %.2f Maxg %.2f Ave %.2f\n", cgen, maxnf, maxf, f, maxgf, avef);
         //displayindfit(indfit, popsize);
         // generate offspring
          for (i=0, ipi= indfit, ipi2=(indfit + popsize); i < popsize; i++, ipi++, ipi2++)
           {
            in = *ipi;
            in2 = *ipi2;
            copyandmutate(genome, ngenes, mutr, prange, (int) *in, (int) *in2);
            in2++;
            in2++;
            *in2 = -9999;
           }
          // store stat data
          *stat = ceval; stat++;
          *stat = maxgf; stat++;
          *stat = maxf; stat++;
          *stat = avef; stat++;
          statn++;
      }

	strcpy(filename, statDir);
     sprintf(filen,"bestgS%d.bgen", seed);
	strcat(filename, "/");
	strcat(filename, filen);
     saveGenotype(filename, bestggeno, ngenes);
	
	 double *gen;
	 int g;
	strcpy(filename, statDir);
	 sprintf(filen,"bestgS%d.gen", seed);
	strcat(filename, "/");
	strcat(filename, filen);
	 fp = fopen(filename,"w!");
	 for(g=0, gen=bestggeno; g < ngenes; g++, gen++)
           fprintf(fp, "%lf ", *gen);
	fclose(fp);

	strcpy(filename, statDir);
     sprintf(filen, "S%d.fit", seed);
	strcat(filename, "/");
	strcat(filename, filen);
     if  ((fp = fopen(filename,"w")) != NULL)
     {
        fprintf(fp, "Gen %d Maxn %.2f Max %.2f g %.2f Maxg %.2f Ave %.2f\n", cgen, maxnf, maxf, f, maxgf, avef);
        fclose(fp);
     }

	strcpy(filename, statDir);
     sprintf(filen, "statS%d.fit", seed);
     strcat(filename, "/");
	strcat(filename, filen);
     if  ((fp = fopen(filename,"w")) != NULL)
         {
            fprintf(fp, "cevaluation bestgfit bestfit avefit \n");
            for(s=0, stat = statistics; s < statn; s++)
              {
               for(ss=0; ss < 4; ss++, stat++)
                 fprintf(fp, "%lf ", *stat);
               fprintf(fp, "\n");
               }
            fclose(fp);
         }

 }


/*
 * save a genotype in a file
 */
void saveGenotype(char* filename, double* genotype, int glen)

{

	FILE *fp;
	
	fp = fopen(filename,"wb!");
	if (!fp)
      printf("ERROR: unable to open file %s\n", filename);
     else
      {
      fwrite(&glen, sizeof(int), 1, fp);
	  fwrite(genotype, sizeof(double)*glen, 1, fp);
      }
	fclose(fp);

}

/*
 * load a genotype from a file
 */
int loadGenotype(char* filename, double* genotype, int glen)

{

	FILE *fp;
	size_t result;
    int np;
	
	fp = fopen(filename,"rb");
	if (!fp)
	 {
	  printf("ERROR: unable to open file %s\n", filename);
	  return(0);
	 }
     else
	 {
      fread(&np, sizeof(int), 1, fp);
      if (np != glen)
        {
         printf("ERROR: the number of parameters contained in the file %s differ from that expected\n", filename);
         //if (!feof(fp) || result != sizeof(double)*glen */)
         //  printf("ERROR: the number of parameters contained in the file %s differ from what is expected\n", filename);
        }
       else
        {
         result = fread(genotype, sizeof(double)*glen, 1, fp);
        }
      fclose(fp);
	 }
    return(1);
}

/*
 * save a genotype in a file
 */
void savePop(char* filename, double **pop, int glen, int popsize)

{

	FILE *fp;
	int p;
	
	fp = fopen(filename,"wb!");
	if (!fp)
      printf("ERROR: unable to open file %s\n", filename);
     else
      {
	  fwrite(&popsize, sizeof(int), 1, fp);
      fwrite(&glen, sizeof(int), 1, fp);
	  for(p=0; p < popsize; p++, pop++)
	    fwrite(*pop, sizeof(double)*glen, 1, fp);
      }
	fclose(fp);

}

/*
 * load a genotype from a file
 */
void loadPop(char* filename, double **pop, int glen, int popsize)

{

	FILE *fp;
    int np;
	int ps;
	int p;
	
	fp = fopen(filename,"rb");
	if (!fp)
	 {
	  printf("ERROR: unable to open file %s\n", filename);
	 }
     else
	 {
	  fread(&ps, sizeof(int), 1, fp);
      fread(&np, sizeof(int), 1, fp);
      if (np != glen || ps != popsize)
        {
         printf("ERROR: the number of parameters contained in the file %s differ from that expected\n", filename);
        }
       else
        {
		 for(p=0; p < popsize; p++, pop++)
           fread(*pop, sizeof(double)*glen, 1, fp);
        }
      fclose(fp);
	 }
}



/*
 * read parameters from the configuration file
 */
bool readEvoParameters(const char* filename)
{
    char *s;
    char buff[1024];
    char name[1024];
    char value[1024];
    char *ptr;
    int section;  // 0=before the section 1=in the section 2= after the section

    section = 0;
    FILE* fp = fopen(filename, "r");
    if (fp != NULL)
    {
        // Read lines
        while (fgets(buff, 1024, fp) != NULL)
        {
            // check whether the relevant section start or end
            if (buff[0] == '[')
            {
              if (section == 1)
                {
                    section = 2;
                    continue;
                }
              if ((section == 0) && (strncmp(buff, "[EVO]",5)==0))
              {
                section = 1;
                continue;
              }
            }

            if (section == 1)
            {

            //Skip blank lines and comments
            if (buff[0] == '\n' || buff[0] == '#' || buff[0] == '/')
            continue;

            //Parse name/value pair from line
            s = strtok(buff, " = ");
            if (s == NULL)
            continue;
            else
            copyandclear(s, name);

            s = strtok(NULL, " = ");
            if (s == NULL)
            continue;
            else
            copyandclear(s, value);

            // Copy into correct entry in parameters struct
            if (strcmp(name, "maxevaluations")==0)
            maxevaluations = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "nreplications")==0)
            nreplications = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "prange")==0)
            prange = strtod(value, &ptr);
            else if (strcmp(name, "mutrate")==0)
            mutrate = strtod(value, &ptr);
            else if (strcmp(name, "stochasticity")==0)
            stochasticity = strtod(value, &ptr);
            else if (strcmp(name, "popsize")==0)
            popsize = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "seed")==0)
            seed = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "steptimelength")==0)
            steptimelength = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "envchangeevery")==0)
            envchangeevery = (int)strtol(value, &ptr, 10);
            else printf("WARNING: Unknown parameter %s in section [EVO] of file %s \n", name, filename);
            }
         }
        fclose (fp);
        if (section == 0)
           printf("WARNING: Missing section [EVO] in file %s \n", filename);
        return(true);
    }
    else
    {
        printf("ERROR: unable to open file %s\n", filename);
        fflush(stdout);
        return(false);
    }
}

/*
 *
 */
void test(int ngenes)
{
}


