/*
 * Evorobot* - A software for running evolutionary robotics experiments
 * Copyright (C) 2005
 * LARAL, Institute of Cognitive Science and Technologies, CNR, Roma, Italy
 * http://laral.istc.cnr.it

 * This program is free software; you
 ls can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#include "main.h"
#include <QTime>

#include <math.h>
#include <locale.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "robot-env.h"
#include "evonet.h"
#include "evoxnes.h"
#include "utilities.h"

extern double evaluate(double* genotype, const int glen, int mode);
extern int initialize(const char* filename);



/*
 * copy the content of the ini file into a file SN.log
 */
void copyIniFile(char* filename, int seed)
{

    char fname[1024];
    char buff[1024];
    FILE* fp1;
    FILE* fp2;

    sprintf(fname, "S%d.log", seed);

    fp1 = fopen(filename, "r");
    fp2 = fopen(fname, "w");
    if (fp1 != NULL)
    {
        // Read lines
        while (fgets(buff, 1024, fp1) != NULL)
        {
          fprintf(fp2, "%s", buff);
        }
        fclose (fp1);
        fclose (fp2);
    }
    else
    {
        printf("ERROR: unable to open file %s\n", filename);
        fflush(stdout);
    }
}


/*
 * describe parameters
 */
void usage()

{
   printf("Evorobot 3.0\n");
   printf(".ini parameter is interpreted as the name of the configuration file\n");
   printf(".bgen or .gen parameter interpreted as the name of a file containing a genotype to be tested\n");
   printf("<number> is interpreted as the seed \n");
   printf("test trigger the execution of the test() function\n");
}


extern void test(int ngenes);

/*
 * Main function
 */
int main(int argc, char *argv[])
{

	 int ngenes;
     char fileini[128];
     char filebgen[128];
     char filegen[128];
     int  pseed;
     bool ufileini;
     bool ufilebgen;
     bool ufilegen;
     bool useed;
     bool utest;
     bool ueval;
	 double f;
     int a;

     // initialize the randomgenerator
     RNG = new RandomGenerator();
	
	 // set USA local conventions
	 setlocale( LC_ALL, "en-US" );
	 setlocale(LC_NUMERIC, "C");

     // parse arguments
     ufileini = false;
     ufilebgen = false;
     ufilegen = false;
     useed = false;
     utest = false;
     ueval = false;
     bestggenotype = NULL;
     for(a=1; a < argc; a++)
     {
        if (filenendwith(argv[a], ".ini"))
        {
           sprintf(fileini, "%s", argv[a]);
           ufileini = true;
        }
        else if (filenendwith(argv[a], ".bgen"))
        {
           sprintf(filebgen, "%s", argv[a]);
           ufilebgen = true;
        }
        else if (argv[a][0] >= 48 && argv[a][0] <= 57)
        {
          pseed = (int)strtol(argv[a], NULL, 10);
          useed = true;
        }
        else if (strcmp(argv[a], "test")==0)
        {
          utest = true;
        }
        else if (strcmp(argv[a], "eval")==0)
        {
          ueval = true;
        }
        else if (filenendwith(argv[a], ".gen"))
        {
           sprintf(filegen, "%s", argv[a]);
           ufilegen = true;
        }
     }
	
	 // display usage information (the first parameter is mandatory)
     if (!ufileini)
     {
         printf("ERROR: you need to specify a .ini file as first parameter\n");
         usage();
     }

	 // initialize
     ngenes = initialize(fileini);
	 if (ngenes <= 0)
	     printf("ERROR: Failure during the initialization process\n");

	 // read EVO parameters
     if (ufileini)
       readEvoParameters(fileini);

     if (useed)
     {
       seed = pseed;
     }

     RNG->setSeed(seed);

     if (ufilebgen)
     {
       bestggenotype = (double *)malloc(ngenes * sizeof(double));
       loadGenotype(argv[2], bestggenotype, ngenes);
     }

     if (ufilegen)
     {
         bestggenotype = (double *)malloc(ngenes * sizeof(double));
         double *gen;
         int g;
         FILE *fp;
         fp = fopen(argv[2],"r");
         for(g=0, gen=bestggenotype; g < ngenes; g++, gen++)
             fscanf(fp, "%lf ", gen);
         fclose(fp);
     }

     if (!utest && !ueval && !ufilebgen && !ufilegen)
	 {
	  for(int r=0; r < nreplications; r++, seed++)
	   {
	    RNG->setSeed(seed);
        evolve(ngenes);
		copyIniFile(fileini, seed);
	   }
       evolve(ngenes);
	 }

     if (utest)
       test(ngenes);

      if (ufilebgen || ufilegen || ueval)
      {
       createInitialStates(0);

       printf("Testing %s with seed %d steplength %d \n", argv[2], seed, steptimelength);
       f = evaluate(bestggenotype, ngenes, 1);
       printf("best g genotype retest %.2f\n", f);
		 
      }

	
}
